/* readtemp.c 


  by Olaf Leidinger - leidola@web.de


  the function query_hddtemp_deamon () 
  comes from gkrellm2's hddtemp plugin.

*/

/*
  released under the terms of the GPL
*/

#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <stdio.h>

#include "readtemp.h"


/************************************************************/
sSensorTemp *new_sSensorTemp(const char *dev, const char *alias, 
			     const char *add, float temp)
{
    sSensorTemp *ret = (sSensorTemp*) malloc (sizeof(sSensorTemp));

    memset (ret, 0, sizeof (sSensorTemp));

    if (dev) 
	/* string passed */
	ret->dev = g_strdup (dev);
    else 
	ret->dev = g_strdup ("");

    if (alias)
	ret->alias = g_strdup (alias);
    else
	ret->alias = g_strdup ("");

    if (add)
	ret->additional = g_strdup (add);
    else
	ret->additional = g_strdup ("");

    ret->temp = temp;

    return ret;
} /* new_sSensorTemp */

/************************************************************/
void free_sSensorTemp(sSensorTemp *sensor)
{
    if (sensor)
    {
	free (sensor->alias);
	free (sensor->dev);
	free (sensor->additional);
	free (sensor);
    }
} /* free_sSensorTemp */

/************************************************************/
int comp_sSensorTemp_str (void* sensor, void* str)
{
    if (sensor && str)
    {
	return strcmp ( S_SENSOR_TEMP(sensor)->dev, str);
    } /* s1 && s2 */
    else
    {
	if (NULL == sensor)
	    return -1;
	else
	    return 1;
    }
} /* comp_sSensorTemp_str */

/************************************************************/
void dump_sSensorTemp (sSensorTemp* sensor, void *reserved)
{
    if (sensor) 
    {
	printf ("%s:\t%.1f C\n", sensor->alias, sensor->temp);
    }
} /* dump_sSensorTemp */

/************************************************************/

GSList *read_cpu_temp_acpi (GSList *s_list, int *state) 
{
    gchar *content = NULL, 
	*p = NULL;
    gsize length   = 0;
    GError *error  = NULL;
    gchar **filenames = NULL, **pf;
    int progress_info[2] = {0, 0};

    const char* sensor_device = DEV_PREFIX_ACPI_CPU DEV_DELIM "cpu_0"; 
/* what about multiple processors? */
    const char* sensor_alias = "CPU 0";
    
    sSensorTemp *sensor = NULL;
    GSList *list;

    /* try reading the content 
       or each file in ACPI_CPU_PATH, one by one; use
       the content of the first file which is readable 
       and parsable.
    */

    filenames = g_strsplit (ACPI_CPU_PATH, ":", 0);

    for (pf=filenames; *pf != 0; pf++)
    {

	if (!g_file_get_contents (*pf,
				   &content, &length,&error))
	{
        g_clear_error(&error);
	    continue; /* could not open the file, try the next one */
	}
    g_clear_error(&error);
	
	progress_info[0]++;    

	/* we have the contents in memory,
	   now parse them */
    
	if (g_ascii_strncasecmp ("temperature:", content, 12) != 0 )
	{
	    /* parse error: temperature expected */
	    g_warning ("parse error: \"temperature:\" expected, %s found in file %s.", 
		       content, *pf);
	    g_free (content);

	    continue; /* parse error; try next file */

	} /* if */
	

	progress_info[1]++;
    
	/* look for the entry  with sensor_device */
    
	list = g_slist_find_custom (s_list, sensor_device, 
				    (GCompareFunc)&comp_sSensorTemp_str);

	if (list) 
	{
	    /* found sensor in the list  - use it */
	    sensor = S_SENSOR_TEMP (list->data);
	}
	else 
	{
	    /* no sensor in the list found; create and append one */
	    sensor = new_sSensorTemp (sensor_device, sensor_alias, 
				      sensor_alias, 0.0f);
	    s_list = g_slist_append (s_list, sensor);
	}
    
	/* now we have (hopefully) a sensor struct which we can write to. */

	p = content + 12; /* points at the char after ':' */
    
	while (g_ascii_isspace (*p) ) /* skip whitespace */
	    p++;
    
	sensor->temp = g_strtod (p, &p); /* read temp from string */
    
    
	while (g_ascii_isspace (*p) ) /* again skip whitespace */
	    p++;  
	
	/* todo: 
	   check what temperature unit is used and convert to kelvin or so
	*/

	g_free (content);
	break;
    } /* for */

    
    if (state) 
    {
	if (progress_info[1])
	    *state = WELL_DONE;
	else
	{
	    if (progress_info[0])
	    {
		/* could not parse any file */
		*state = ERROR_PARSE;
	    }
	    else
	    {
		fprintf(stderr,"acpi_temperature: could not read any of %s.\n",
			ACPI_CPU_PATH);	
		*state = ERROR_READING;	
	    }
		
	}
    } /* if state */
    g_strfreev (filenames);

    return s_list;
} /* read_cpu_temp_acpi */

/************************************************************/


GSList* read_hdd_temp (GSList *s_list, int *state) 
{
    gchar *content = NULL, *host = HDD_TEMP_HOST_DEFAULT,
	*p = NULL;
    sSensorTemp *sensor = NULL;
    gchar **strv, **pv, *dev;
    gint port = HDD_TEMP_PORT_DEFAULT;

    /* TODO:
       add support for remote queries 
    */
  p = content = query_hddtemp_deamon (host, port);

  if (!content)
  {
      /* some error while reading must have occured */
      if (state) *state = ERROR_READING;
      return s_list;
  }
   

  /* the string has the form:
     |device|name|temp|unit|
  */

  /* perhaps it'd be better to write
     a parser instead of splitting it into
     substrings; at least this would consume less memory.
  */
  
  while (*p != '|') /* find the beginning of the information */
      p++;

  pv = strv = g_strsplit (p+1, "|", 0); /* split string to tokens */

  /*
    pv[0 + 5*n] unix-device
    pv[1 + 5*n] disk name
    pv[2 + 5*n] temperature
    pv[3 + 5*n] unit
    pv[4 + 5*n] empty
  */

  while (*pv) 
  {
      gchar *buffer = NULL;
      GSList *list  = NULL;
      
      dev = g_strdup_printf ("%s" DEV_DELIM "%s:%i\%s", 
			     DEV_PREFIX_HDDTEMP, host, port,
			     g_strdelimit (pv[0], "/", '\\'));

      /* look for the entry  with name buffer */
    
      list = g_slist_find_custom (s_list, dev, 
				  (GCompareFunc)&comp_sSensorTemp_str);

      if (list) 
      {
	  /* found sensor in the list  - use it */
	  sensor = S_SENSOR_TEMP (list->data);
      }
      else 
      {
	  /* no sensor in the list found; create and append one */

	  sensor = new_sSensorTemp (dev, pv[1], pv[1], 0.0f);
	  s_list = g_slist_append (s_list, sensor);
      }
      
      sensor->temp = g_strtod (pv[2],NULL); /* read temp from string */

      pv += 5;

      free (buffer);   /* TODO: find out why */
      g_free (dev);
  } /* while *pv */


  free (content);
  g_strfreev (strv);

  return s_list;
} /* read_hdd_temp */    

/************************************************************/
char * query_hddtemp_deamon(const char * server, unsigned int port) 
{
    /* from gkrellm2 hddtemp plugin */

    struct sockaddr_in  saddr;
    struct hostent      *he;
    int                 sk;
    int                 n, pos;
    const char          *serv_name = server;
    char                buff[2];
    
    static char         *str = NULL;
    static unsigned int str_size = 0;
  
    if((sk = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
	perror("query_hddtemp_deamon: socket");
	return NULL;
    }
    
    memset(&saddr, 0, sizeof(struct sockaddr_in));
    if((he = gethostbyname(serv_name)) == NULL) 
    {
	perror("query_hddtemp_deamon: gethostbyaddr");
	return NULL;
    }

    memcpy(&(saddr.sin_addr.s_addr), he->h_addr, he->h_length);
    
    saddr.sin_family = AF_INET;
    saddr.sin_port = htons(port);
    
    if(connect(sk, (struct sockaddr *)&saddr, (socklen_t)sizeof(saddr)) == -1) {
	perror("query_hddtemp_deamon: connect");
	return NULL;
    }

    if(str == NULL) {
	str_size = 2;
	
	if((str = (char*)malloc(str_size * sizeof(char))) == NULL) {
	    perror("query_hddtemp_deamon: malloc");
	    return NULL;
	}
    }
    
    pos = 0;
    while((n=read(sk, &buff, sizeof(buff))) > 0) {
	if(pos+n+1 > str_size) {
	    str_size *= 2;
	    if((str=(char*)realloc(str, str_size)) == NULL) {
		perror("query_hddtemp_deamon: realloc");
		return NULL;
	    }
	}
	
	strncpy(str+pos, buff, n);
	pos += n;
    }
    str[pos] = '\0';  
    
    close(sk);
    
    return (char*) strdup(str);
}

/************************************************************/

GSList *read_lm_sensor_temp (GSList *s_list, int *state)
{
    GDir *dir  = NULL;
    char *path = LINUX_2_6_SYSFS_I2C_PATH;
    GError *error = NULL;
    gchar *dname, *full_path; /* directory name in dir */
    sSensorTemp *sensor = NULL;      

    dir = g_dir_open (path, 0, &error); 
    
    if (!dir) 
    {
	fprintf (stderr, "lm_sensor_temp: error while reading %s\n"
		 "Do you use kernel 2.6 and have lm_sensor modules loaded?\n", 
		 path);
        g_clear_error(&error);
	*state = ERROR_READING;
	return s_list;
    }
    g_clear_error(&error);
	
    /* look for i2c-? */

    while ( (dname = (gchar*) g_dir_read_name (dir)) )
    {
	/* as long as there are files in dir */
	full_path = g_strdup_printf ("%s/%s", path, dname);

	if (g_file_test (full_path, G_FILE_TEST_IS_DIR))
	{
	    /* we have a directory */

	    if (g_str_has_prefix ( dname, LINUX_2_6_I2C_PREFIX  ) )
	    {
		/* okay, look for the devices */
		GDir *subdir;
		gchar *full_spath, *subdname;
		char *i2c_prefix = dname + strlen ( LINUX_2_6_I2C_PREFIX );
		/* this should be only a number */


		/* now look for directories beginning with i2c_prefix */
		
		subdir = g_dir_open (full_path, 0, &error); 
    
		if (!subdir) 
		{
		    fprintf (stderr, "lm_sensor_temp: internal error?\n"
			     "  could not open %s\n", full_path);
		    g_free (full_path);
            g_clear_error(&error);

		    return s_list;
		}
        g_clear_error(&error);

		while ( (subdname = (gchar*) g_dir_read_name (subdir)) )
		{
		    /* as long as there are files in subdir */
		    full_spath = g_strdup_printf ("%s/%s",full_path, subdname);
		    
		    if (g_file_test (full_spath, G_FILE_TEST_IS_DIR)) 
		    {
			/* we have a directory */
			
			if (g_str_has_prefix ( subdname, i2c_prefix) )
			{
			    /* now we have found an potential temp-sensor */
			    gchar *full_sensor_path = NULL;
			    gchar *content, *dev;
			    gint i, size;

			    for (i=1; i < G_MAXINT; i++)
			    {
				full_sensor_path = 
				    g_strdup_printf ("%s/temp%i_input",
						     full_spath, i);

				if ( g_file_test (full_sensor_path,
						  G_FILE_TEST_EXISTS))
				{
				    /* now we should have a temperature */
				    if ( (g_file_get_contents (full_sensor_path, 
							       &content, &size, &error) ) )
				    {
					/* temperature in content - parse it */
					/* on kernel 2.6 its basically the temp in celsius * 1000 */

					/* look for the entry  with the same device */
    
					GSList *list;
					dev = g_strdup_printf ("%s" DEV_DELIM "%s", 
							       DEV_PREFIX_LM_SENSORS, 
							       g_strdelimit (full_sensor_path + LINUX_2_6_SYSFS_I2C_PATH_LEN+1, 
									     "/", '\\'));

					list = g_slist_find_custom (s_list, dev, 
								    (GCompareFunc)&comp_sSensorTemp_str);

					if (list) 
					{
					    /* found sensor in the list  - use it */
					    sensor = S_SENSOR_TEMP (list->data);
					}
					else 
					{
					    /* no sensor in the list found; create and append one */
					    
					    sensor = new_sSensorTemp (dev, full_sensor_path+strlen(full_spath)+1, NULL, 0.0f);
					    s_list = g_slist_append (s_list, sensor);
					}
					
					sensor->temp = g_strtod (content,NULL)/1000.0; /* read temp from string */

					g_free (content);
					g_free (dev);
					
				    } /* if got content */
                    g_clear_error(&error);
				    
				} /* if file exists */
				else
				{
				    /* this is the first assumed sensor
				       that doesn't exist; break here */
				    g_free (full_sensor_path);
				    break;
				}

				g_free (full_sensor_path);
			    } /* for */
			} /* if has prefix */
		    } /* if is directory */

		    g_free (full_spath);
		} /* while subdname */

		g_dir_close (subdir);		
	    } /* if has prefix */

	} /* if directory */

	g_free (full_path);
    } /* while */

    g_dir_close (dir);

    *state = WELL_DONE;

    return s_list;
} /* read_lm_sensor_temp */
